# $Id: dglib.pm,v 1.5.2.16 2013/03/07 12:13:39 dsathe Exp $ #
#ident "@(#)CSODG:$RCSfile: dglib.pm,v $	$Revision: 1.5.2.16 $"
#
# $Copyright: Copyright (c) 2016 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
eval 'exec /opt/VRTSperl/bin/perl -Sw -I${VCS_HOME}/lib $0 ${1+"$@"}'
	if 0;

package dglib;
use strict;
use ag_i18n_inc;

use lib qw(.);
use dgconfig () ;
use vars qw(%CONF);
*CONF = \%dgconfig::CONF;
my $SU = $CONF{SU};
my $PASSWDFILE = $CONF{PASSWDFILE};

my $LockFilePath = "$VCSHOME" . "/bin/OraDG/.";
my $LockFileExt = "_LCK.dont_remove";

my $RemoteNode = "";

################# FUNCTION DEFN START ##########################
# Function : check_oracle_home
# Desc     : Checks for inexistence of ORACLE_HOME 
# Input    : $Home 
# Output   : None
# Returns  : (1) -  home does not exist  
#
sub check_oracle_home  {

	my ($Home) =@_;
	unless ( -d $Home ) {
		return  1;
	}
	return 0;
}

# Function : check_oracle_client
# Desc     : Checks if svrmgrl/sqlplus does not exist .
# Desc     : If they exist , this function sets DBAMGR and CONNECTSTR
# Input    : $Home  
# Output   : $DBAMGR,$CONNECTSTR - changes values of these variables 
# Returns  : (1) - no suitable client exists 
#
sub check_oracle_client {
my ($Home,$DBAMGR,$CONNECTSTR)=@_;

	if ( -x "$Home/bin/sqlplus" ) {
		#
		# logon using OS authentication.
		#
		$DBAMGR = "$Home/bin/sqlplus -s /nolog";
		$CONNECTSTR = "connect / as sysdba";
		$_[1] = $DBAMGR;
		$_[2] = $CONNECTSTR;
	} else {
		return 1 ;
	}
	return 0;
}

# Function : check_srvctl
# Desc     : Checks if srvctl does not exist (RAC installations) .
# Desc     : If they exist , this function sets DBAMGR 
# Input    : $Home  
# Output   : $DBAMGR - changes values of these variables 
# Returns  : (1) - no suitable client exists 
#
sub check_srvctl {
my ($Home,$DBAMGR)=@_;

	if ( -x "$Home/bin/srvctl" ) {
		$DBAMGR = "$Home/bin/srvctl";
		$_[1] = $DBAMGR;
	}   else {
		return 1 ;
	}
	return 0;
}


# Function : get_db_state
# Desc     : Retrieve database state information from V$DATABASE
# Input    : $Owner - owner of executables
#          : $LD_ENV - Env var string returned from function "check_environment"
#          : $DBAMGR - client utility
#          : $CONNECTSTR - client utility specific connectstring
# Output   : $DGRole 
#          : $DGStatus 
#          : $OpenMode 
# Returns  : $sqlout
#
sub get_db_state() {
	my($Owner, $LD_ENV, $DBAMGR, $CONNECTSTR, $DGRole, $DGStatus, $OpenMode) = @_;
	my ($SQLStmt, $sqlout);
	my (@lines, $i);

	$_[4] = "";
        $_[5] = "";
        $_[6] = "";
	
	$SQLStmt = "SELECT DATABASE_ROLE FROM V\\\$DATABASE;\n";
	$SQLStmt .= "SELECT SWITCHOVER_STATUS FROM V\\\$DATABASE;\n";
	$SQLStmt .= "SELECT OPEN_MODE FROM V\\\$DATABASE;\n";

	$sqlout = `$SU $Owner -c "$LD_ENV; echo LD_LIBRARY_PATH - \\\$LD_LIBRARY_PATH; $DBAMGR" << !
$CONNECTSTR
$SQLStmt
!`;
	if ($sqlout =~ /\n(ORA-\d+.*$)/mi) {
                #Error in executing SQL query. return exact error.
		return $1;
        }

	@lines = split(/\n+/, $sqlout);

	for ($i=0; $i < @lines; $i++) {
		if ($lines[$i] =~ /DATABASE_ROLE/i) {
			chomp($DGRole = $lines[$i + 2]);
			$i += 2;
		} elsif ($lines[$i] =~ /SWITCHOVER_STATUS/i) {
			chomp($DGStatus = $lines[$i + 2]);
			$i += 2;
		} elsif ($lines[$i] =~ /OPEN_MODE/i) {
			chomp($OpenMode = $lines[$i + 2]);
			$i += 2;
		}
	}
	$_[4] = $DGRole;
	$_[5] = $DGStatus;
	$_[6] = $OpenMode;
	return $sqlout;
}

# Function : get_link_res_attr
# Desc     : Retrieve the instance attributes from Oracle resource
# Input    : $LinkRes - the resource name of the local DG link resource
# Input    : $NodeName - the resource name of the local DG link resource
# Output   : $Sid
# Output   : $Owner
# Output   : $Home
# Output   : $EnvFile - source file for env vars
# Returns  : 1 - the associated OraDGLink resource does not exist
#
sub get_link_res_attr {
	my ($LinkRes, $NodeName, $Sid, $Owner, $Home, $EnvFile) = @_;
	my $i = 2;
	for my $ResAtt ("Sid","Owner","Home","EnvFile") {
		chomp($_[$i] = `$VCSHOME/bin/hares -value $LinkRes $ResAtt $NodeName`);
		$i += 1;
	}
	return 1 if $_[2] eq "" || $_[3] eq "" || $_[4] eq "";
	return 0;
}

# Function : get_local_nodename
# Desc     : Retrieve the name of the local node
# Input    : None
# Output   : None
# Returns  : Name of the local node

sub get_local_nodename {
	my $SysName = "${VCSCONF}/conf/sysname";
	my $GlobZoneName = "$VCSCONF/.vcshost";
	my $LocSys;

	if ( -r "$SysName" ) {
		open(SNAME, "<$SysName");
	} elsif ( -r "$GlobZoneName" ) {
		open(SNAME, "<$GlobZoneName");
	} else {
		open(SNAME, "uname -n |");
	}
	chomp($LocSys = <SNAME>);
	close SNAME;
	return $LocSys;
}

# Function : get_remote_info
# Desc     : Retrieve the name of the remote node and cluster
# Input    : $ResID
# Output   : $RemClust
# Output   : $RemNode
# Returns  : Name of the local node

sub get_remote_info {
        my ($ResID, $RemClus, $RemNode) = @_;
	$RemClus = $RemNode = "";
        if (&dglib::is_global_group($ResID)) {
        	open(RDSP, "$VCSHOME/bin/hares -display $ResID |");
                while (<RDSP>) {
			next unless /^$ResID\s+State\s/;
                	next if /\slocalclus/;
                	my @line = split();
                	($RemClus, $RemNode) = split(':', $line[2]);
                	my $ret = `$HARES -action $ResID DGStatus -sys $RemNode -clust $RemClus 2>&1`;
                	last if $ret =~ /DATABASE_ROLE/i;
                	$RemClus = $RemNode = "";	
                }
        } else {
		my $OraRes = `$VCSHOME/bin/hares -value $ResID LinkRes`;
                chomp $OraRes;
		my $LocSys =&dglib::get_local_nodename();
                open(RDSP, "$VCSHOME/bin/hares -display $OraRes |");
                while (<RDSP>) {
                        next unless /^$OraRes\s+State\s/;
                        next if /\sOFFLINE/;
                        my @line = split();
                        $RemNode = $line[2];
                        if ($RemNode eq $LocSys) {
                                $RemNode = "";
                                next;
                        }
                        my $ret = `$VCSHOME/bin/hares -action $ResID DGStatus -sys $RemNode 2>&1`;
                        last if $ret =~ /DATABASE_ROLE/i;
                        $RemClus = $RemNode = "";
                }
	}
        # Make sure the pipe is empty
        while (<RDSP>) {next;}
        close RDSP;
        return 1 if ( $RemClus eq "" && $RemNode eq "" );
        $_[1] = $RemClus;
        $_[2] = $RemNode;
        return 0;
}

# Function : is_global_group
# Desc     : check if configured group is global group
# Input    : $ResID
# Returns  : 0 - not a global group(RDC)
# Returns  : 1 - global group

sub is_global_group {
        my ($ResID) =  @_;
        my $ret = 0;
        my $cmd = "$VCSHOME/bin/hares -value $ResID Group";
        my $Grp = `$cmd`;
        chomp $Grp;
        $cmd = "$VCSHOME/bin/hagrp -value $Grp  ClusterList";
        my $ClusterList = `$cmd`;
        chomp $ClusterList;

        if ( $ClusterList eq "" ) {
                VCSAG_LOG_MSG("I", "Remote Copy Service Group $Grp is not configured as global group.", 56, "$Grp");
                $ret = 0;
        } else {
                VCSAG_LOG_MSG("I", "RemoteCopy Service Group $Grp is configured as global group.", 57, "$Grp");
                $ret = 1;
        }
        return $ret;
}


# Function : lock_create
# Desc     : Creates a lock file to mark the resource as online
# Input    : $ResID
# Returns  : 0 - success
#
sub lock_create {
my ($ResID)=@_;

	my $LockFile = "$LockFilePath" . "$ResID" . "$LockFileExt";
	open(LCK, ">$LockFile") or return 1;
	close LCK;
	return 0;
}

# Function : lock_exists
# Desc     : Checks if the lock file for a resource exists
# Input    : $ResID
# Returns  : 0 - file exists
# Returns  : 1 - file does not exists
#
sub lock_exists {
my ($ResID)=@_;

	my $LockFile = "$LockFilePath" . "$ResID" . "$LockFileExt";
	return 0 if -f $LockFile;
	return 1;
}

# Function : lock_remove
# Desc     : Removes the lock file for a resource to mark it offline
# Input    : $ResID
# Returns  : 0 - success
#
sub lock_remove {
my ($ResID)=@_;

	my $LockFile = "$LockFilePath" . "$ResID" . "$LockFileExt";
	unlink("$LockFile");
	return 0;
}

# Function : kill_open_sessions
# Desc     : Kills open sessions on the standby prior to role swap
#          : Only SYS user should be able to connect to the standby
# Input    : $Owner - owner of executables
#          : $LD_ENV - Env var string returned from function "check_environment"
#          : $DBAMGR - client utility
#          : $CONNECTSTR - client utility specific connectstring
# Returns  : 0 - success
#
sub kill_open_sessions {
	my($Owner, $LD_ENV,$DBAMGR,$CONNECTSTR) = @_;
	my (@lines, $i);
	my ($SidID, $Serial);
	my $KillStr = "";
	my $OutStart = 0;
	my $SQLStmt = "SELECT SID, SERIAL# FROM V\\\$SESSION WHERE USERNAME = \'SYS\';\n";

	my $sqlout = `$SU $Owner -c "$LD_ENV; echo LD_LIBRARY_PATH - \\\$LD_LIBRARY_PATH; $DBAMGR" << !
$CONNECTSTR
$SQLStmt
!`;

	@lines = split(/\n+/, $sqlout);

	for ($i=0; $i < @lines; $i++) {
		next if $lines[$i] =~ /SID/i;
		next if $lines[$i] =~ /^\s+$/i;
		if ($lines[$i] =~ /----/) {
			$OutStart = 1;
			next;
		}
		next unless $OutStart;
		chomp;
		$lines[$i] =~ s/^\s+//;
		($SidID, $Serial) = split(/\s+/, $lines[$i]);
		if ($SidID >= 0 && $Serial >= 0) {
			$KillStr .= "ALTER SYSTEM KILL SESSION \'$SidID,$Serial\';\n";
		}
	}
	if ($KillStr ne "") {
		VCSAG_LOG_MSG("I", "KillStr: $KillStr", 21, "$KillStr");

		$sqlout = `$SU $Owner -c "$LD_ENV; echo LD_LIBRARY_PATH - \\\$LD_LIBRARY_PATH; $DBAMGR" << !
$CONNECTSTR
$KillStr
!`;

	}
	return 0;
}

# Function : set_instance_env
# Desc     : Checks the EnvFile, if not valide sets $ValidFile to 1
# Desc     : Else sets the Oracle related env vars 
# Input    : $EnvFile - source file for env vars
# Input    : $Owner 
# Input    : $Home 
# Output   : $ValidFile - 
# Returns  :  $LD_ENV - Env vars in format to be set 
# Note     :  This function sets LD_LIBRARY_PATH ( and equivalents) 
#          :  TWO_TASK, extra env vars 
#
sub set_instance_env {
	my ($EnvFile, $Owner, $Home, $ValidFile) = @_;
	my $UseCSH = 0;
	my ($LD_ENV);
	my $EnvFileExists = 0;

	open (PASSWD, $PASSWDFILE);
	while (<PASSWD>) {
		next unless /^$Owner:/;
		$UseCSH = 1 if /csh$/;
		last;
	}
	# Make sure the pipe is empty
	while (<PASSWD>) {next;}
	close PASSWD;

	my $EXTRA_ENV= dgconfig::set_extra_env($UseCSH, $Home);
	if ( $EnvFile ) {
		
		#
		# It is the user's responsibility to ensure that EnvFile can
		# be sourced successfully using the login shell of $Usere 
		#
		# First ensure that the file is a Plain file AND a Text file.
		# Other file types (e.g. Named pipe) etc. cause $EnvFile to be
		# NOT sourced, with a message in the engine log saying so.
		#
		if ( -f $EnvFile && -T $EnvFile ) {
			#
			# Doing a 'test -r' on the file as user '$Owner' ensures that the 
			# permission are correct for user '$Owner'.
			# Execute permissions are NOT needed for $EnvFile to source it.
			#
			# Notes:
			#   * `` always invokes sh.
			#
			my $readerr = 0;
			$readerr = `$SU $Owner -c "test -r $EnvFile; echo $?" 2>&1`;
			if ( $readerr ) {
				$EnvFileExists = 1;	# read permission exists
				$_[3] = 0;		#is a valid file
			} else {
				$EnvFileExists = 0;	# error encountered by test -r
				$_[3]=1; 		# Not a valid file
			}
		} else {
			$EnvFileExists = 0;		# not a text file
			$_[3]=1; 			# Not a valid file
		}
	}
	# Important points from the man page of 'su':
	#   * Regarding LD_LIBRARY_PATH,
	#       Variables with LD_ prefix are removed for security  reasons.
	#       Thus,  su  bin will not retain previously exported variables
	#       with LD_ prefix while becoming user bin.
	#   * The environment  of  user root is passed along to user oracle,
	#       with the exception of $PATH, which is controlled by PATH and
	#       SUPATH in /etc/default/su.
	#   * $LANG environment variable: The su should work even if LANG=ja:
	#       If any  of  the  LC_*  variables  (  LC_CTYPE,  LC_MESSAGES,
	#       LC_TIME,   LC_COLLATE,  LC_NUMERIC,  and  LC_MONETARY)  (see
	#       environ(5)) are not set in the environment, the  operational
	#       behavior  of  su  for  each corresponding locale category is
	#       determined by the value of the LANG environment variable. If
	#       LC_ALL  is  set,  its contents are used to override both the
	#       LANG and the other LC_* variables.  If  none  of  the  above
	#       variables  are  set in the environment, the "C" (U.S. style)
	#       locale determines how su behaves.

	# NOTE: that the LD_LIBRARY_PATH should be set AFTER the su to $Owner,
	#	and NOT for environment of the user (root) running this script.
	#
	#	Be careful with the backslashes in this script.
	#
	my ($ldlibpath);
	if ( -d "$Home/lib64" ) {
		 $ldlibpath="/usr/lib:$Home/lib64:\\\$LD_LIBRARY_PATH";
	} else {
		$ldlibpath="/usr/lib:$Home/lib:\\\$LD_LIBRARY_PATH";
	}
	
	# If CSH  then setenv else export environment variables 
	if ($UseCSH) {
		$LD_ENV = "setenv LD_LIBRARY_PATH \\\"\\\"; ". $EXTRA_ENV;
		#
		# Note the .= concatenation operator below
		#
		if ($EnvFileExists) {
			$LD_ENV .= "source $EnvFile; ";
		}
		$LD_ENV .= "setenv LD_LIBRARY_PATH $ldlibpath;unsetenv TWO_TASK";
	} else {
		$LD_ENV = "LD_LIBRARY_PATH=\"\"; export LD_LIBRARY_PATH; ". $EXTRA_ENV;
		#
		# Note the .= concatenation operator below
		#
		if ($EnvFileExists) {
			$LD_ENV .= ". $EnvFile; ";
		}
		$LD_ENV .= "LD_LIBRARY_PATH=$ldlibpath; export LD_LIBRARY_PATH;TWO_TASK=; export TWO_TASK";
	}
	return $LD_ENV;
}

# Function : runsql
# Desc     : Execute SQLSTATEMENTS using client DBAMGR
# Input    : $Owner - owner of executables
#          : $LD_ENV - Env var string returned from function "check_environment"
#          : $DBAMGR - client utility
#          : $CONNECTSTR - client utility specific connectstring
#          : $TOLOG - (1) - the message has to be formatted with BEGLINE and ENDLINE
#          : $SQLSTATEMENTS - statements to execute
# Output   : None 
# Returns  : $out
#
sub runsql() {
	my($Owner, $LD_ENV, $DBAMGR, $CONNECTSTR, $TOLOG, $SQLSTATEMENTS) = @_;

	my $OH=$ENV{"ORACLE_HOME"};
	my $OS=$ENV{"ORACLE_SID"};
	my $BEGLINE="\n+-----------------------------------------------------------" .
                                                        "---------+\n";
	my $ENDLINE="+===========================================================" .
                                                        "=========+";
	my $out = "";

	VCSAG_LOG_MSG("I", "SQL statements executing: $SQLSTATEMENTS", 21, "$SQLSTATEMENTS");

	if (${SQLSTATEMENTS}) {
		
		$out = `$SU $Owner -c "$LD_ENV; echo LD_LIBRARY_PATH - \\\$LD_LIBRARY_PATH; $DBAMGR" << !
${CONNECTSTR}
${SQLSTATEMENTS}
!`;
	}
	
	#Format messages if set.
	if (${TOLOG}) {
		$out = $BEGLINE . $out . $ENDLINE; 
	}

	return $out;

} # End of sub runrsql()

# Function : get_oracle_error
# Desc     : Collect oracle error if any during executing DGDemotePri action entry point
# Input    : resource id
# Output   : None
# Returns  : $oraerrout
#
sub get_oracle_error
{
        my ($resid) = @_;
        my ($oraerrout, $oraerrfilepath);

        $oraerrfilepath = "$VCSHOME" . "/bin/OraDG/." . "$resid" . "_oraerr";
        if ( -e "$oraerrfilepath" ) {
		local $/=undef;
                open ( FILDES, $oraerrfilepath ) || return 1;
		binmode FILDES;
                $oraerrout  = <FILDES>;
                close ( FILDES ) || return 2;

                unlink ( $oraerrfilepath );
		if ( ! ($oraerrout =~ /ORA-/i) ){
			$oraerrout = "";
		}
                return $oraerrout;
        }

        return "";
} # End of sub get_oracle_error()


# Function : set_oracle_error
# Desc     : Dump oracle error if any during executing DGDemotePri action entry point
# Input    : resource id and oracle error output
# Returns  : > 0 if any error, otherwise 0
#
sub set_oracle_error
{
        my ($resid, $oraerrout) = @_;
        my ($oraerrfilepath);

        $oraerrfilepath = "$VCSHOME" . "/bin/OraDG/." . "$resid" . "_oraerr";
        if ( -e "$oraerrfilepath" ) {
		VCSAG_LOG_MSG("I", "unlinking oraerrfilepath file if already exists", 21);
                unlink ( $oraerrfilepath ) || return 1;
        }

	if ( $oraerrout =~ /ORA-/i ){
        	open ( FILDES, ">$oraerrfilepath" ) || return 2;
        	print FILDES $oraerrout;
        	close ( FILDES ) || return 3;
	}
        return 0;
} # End of sub set_oracle_error()

# Function : is_flashback_on
# Desc     : It is used to check whether flashback is enabled or not.
# Input    : $Owner - owner of executables
#          : $LD_ENV - Env var string returned from function "check_environment"
#          : $DBAMGR - client utility
#          : $CONNECTSTR - client utility specific connectstring
# Returns  :  1 if flashback is on, 2 if oracle is not opened, otherwise 0.
#
sub is_flashback_on
{
	my ($Owner, $LD_ENV, $DBAMGR, $CONNECTSTR) = @_;
	my ($SQLStmt, $sqlout);

	$SQLStmt = "select flashback_on from v\\\$database;\n";
	$sqlout = &dglib::runsql($Owner,$LD_ENV,$DBAMGR,$CONNECTSTR,0,$SQLStmt);


	if ($sqlout =~ /ORA-/i) {
		VCSAG_LOG_MSG("I","sqlout: $sqlout", 66, "$sqlout");
		return 2;
	} elsif ($sqlout =~ /YES/i) {
        	return 1;
	} 
	
	return 0;
}

# Function : get_oracle_version
# Desc     : It is used to find the version of Oracle.
# Input    : $version - passed as reference. Oracle version will be copied.
#          : $release - passed as reference. Oracle release will be copied.
# Returns  : always 1
#
sub get_oracle_version
{
        my ($version, $release) = @_;
        my ($out, @out1);
        my ($lineval, $string, $str);

        my $Home = $ENV{'ORACLE_HOME'};
        $out = `$Home/bin/sqlplus -V`;
        @out1 = split(/\n/,$out);
        foreach $lineval (@out1) {
                chomp $lineval;
                if($lineval=~/Release/) {
                        $string = $lineval;
                        ($str = $string) =~ s/.*Release\s+(\d+)\.(\d+).*/$1/;
                        $release = $2;
                        last;
                }
        }

        $_[0] = $str;
        $_[1] = $release;

        return 1;
}


# Function : flashback_recovery
# Desc     : It is used to recover old failed primary database and converting to physical standby.
# Input    : $Owner - owner of executables
#          : $LD_ENV - Env var string returned from function "check_environment"
#          : $DBAMGR - client utility
#          : $CONNECTSTR - client utility specific connectstring
#          : $ResID - value of the dataguard resource
#          : $LinkRes - oracle resource name
#          : $LockSys - local system node
#          : $LockRole - database role on local system node
# Returns  : 0 - Flashback is not on on database
#	   : 1 - Recovered database using flashback successfully
#	   : 2 - Flashback failed, exiting without doing anything.
#
sub flashback_recovery
{
	my ($Owner, $LD_ENV, $DBAMGR, $CONNECTSTR, $ResID, $LinkRes, $LocSys, $LocRole) = @_;
	my ($RemClus, $RemNode);
	my ($SQLStmt, $sqlout);
	my $ret = 0;

	$ret = &dglib::is_flashback_on ($Owner, $LD_ENV, $DBAMGR, $CONNECTSTR);

	if ( $ret == 1 ) {
		VCSAG_LOG_MSG("I","FLASHBACK is ON, hence checking whether flashback recovery is required or not.", 54);

		my ($grp) = split('\n', `$HARES -value $ResID Group`);
		my ($auth) = split('\n',`$HAGRP -value $grp Authority`);
		if ($auth) {
			VCSAG_LOG_MSG("E","Node $LocSys does not have authority to do flashback recovery.", 74, "$LocSys");
			print "Current node is not authorised to do flashback recovery.";
			exit 1;
		}

                $ret = &dglib::get_remote_info($ResID, $RemClus, $RemNode);
                if ($ret) {
			VCSAG_LOG_MSG("I","Normal online as primary database is already open and STANDBY is down.", 55);
                	# Do nothing, primary database is already open and STANDBY is down.
                } else {
			#set -sys and -clust attribute to work with RDC and gco
			if ($RemClus eq "" ) {
				$RemoteNode = "-sys $RemNode";
			} else {
				$RemoteNode = "-sys $RemNode -clust $RemClus";  
			}
			
			my $RemDGStatus = `$HARES -action $ResID DGStatus -sys $RemoteNode 2>&1`;
			my ($RemRole, $mode, $status, $i);
                        my @lines = split('\n', $RemDGStatus);
                        for ($i=0; $i < @lines; $i++) {
                        	if ($lines[$i] =~ /DATABASE_ROLE/i) {
                                	chomp($RemRole = $lines[$i + 2]);
                                        $i += 2;
                                } elsif ($lines[$i] =~ /SWITCHOVER_STATUS/i) {
                                        chomp($status = $lines[$i + 2]);
                                        $i += 2;
                                } elsif ($lines[$i] =~ /OPEN_MODE/i) {
                                        chomp($mode = $lines[$i + 2]);
                                        $i += 2;
                                }
                         }
                         
			VCSAG_LOG_MSG("I", "Oracle dataguard info on old physical standby database node $RemNode: role = $RemRole, mode = $mode, status = $status.", 56, "$RemNode", "$RemRole", "$mode", "$status");
			my $status_flag="";
			my ($version, $release);
			my $ret = &dglib::get_oracle_version ($version, $release);
			if ( $version == 11 && $release == 2) {
				$status_flag = "FAILED DESTINATION";
			} else {
				$status_flag = "NOT ALLOWED";
			}

                        if ( ($LocRole =~ /PRIMARY/i) && ($RemRole =~ /PRIMARY/i)) {
				# Determining the SCN at which the old standby database became the primary database to apply on failed PRIMARY for flashback.
				my @SCNout = split ('\n', `$HARES -action $ResID getremoteSCN -sys $RemoteNode 2>&1`);
				my $SCN = $SCNout[1];
				VCSAG_LOG_MSG("I","SCN calculated on old physical standby database is $SCN", 57, "$SCN");
                                if ($SCN) {
					# checking the open mode of the database
					# proceed for recovery if it is mounted
					$SQLStmt = "SELECT OPEN_MODE FROM V\\\$DATABASE;\n";
                                	$sqlout = &dglib::runsql($Owner,$LD_ENV,$DBAMGR,$CONNECTSTR,0,$SQLStmt);
					if ( !($sqlout =~ /MOUNT/i) ) {
						# Starting Flashback on the failed PRIMARY database.
						system("$HARES -offline $LinkRes -sys $LocSys");
			                	system("$HARES -wait $LinkRes State offline -sys $LocSys");
			                	system("$HARES -online $LinkRes -sys $LocSys");
			                	system("$HARES -wait $LinkRes State online -sys $LocSys");
		                       	}

					$SQLStmt = "FLASHBACK DATABASE TO SCN $SCN;\n";
                                	$sqlout = &dglib::runsql($Owner,$LD_ENV,$DBAMGR,$CONNECTSTR,0,$SQLStmt);
					if ( $sqlout =~ /ORA-/i) {
						VCSAG_LOG_MSG("E","Unable to flashback database to SCN $SCN, error: $sqlout", 47, "$SCN", "$sqlout");
						exit 1;
					}
		                       
					 
					# Convering the failed Old PRIMARY database to PHYSICAL database.
					VCSAG_LOG_MSG("I","Convering the failed Old PRIMARY database to PHYSICAL database.", 58);
					$SQLStmt = "ALTER DATABASE CONVERT TO PHYSICAL STANDBY;\n";
                                        $sqlout = &dglib::runsql($Owner,$LD_ENV,$DBAMGR,$CONNECTSTR,0,$SQLStmt);
                                        if ( $sqlout =~ /ORA-/i) {
                                                VCSAG_LOG_MSG("E","Unable to convert physical standby, error: $sqlout", 50, "$sqlout");
                                                exit 1;
                                        }
					
					system("$HARES -offline $LinkRes -sys $LocSys");
			                system("$HARES -wait $LinkRes State offline -sys $LocSys");
			                system("$HARES -online $LinkRes -sys $LocSys");
			                system("$HARES -wait $LinkRes State online -sys $LocSys");
					VCSAG_LOG_MSG("W","Failed PRIMARY converted to PHYSICAL STANDBY. Administrator intervention is required to resolve gap.", 60);
					return 1;
                                }
				return 2;
                        } else {
				return 3;
			}
		}
	} elsif ( $ret == 2) {
		VCSAG_LOG_MSG("E","Oracle is not available, hence can't proceed for flashback recovery. Please run flashbackRecover action entry point to recover once Oracle is available.", 71);
		exit 4;
	} else {
		VCSAG_LOG_MSG("I","FLASHBACK is not ENABLED, hence can't proceed with flashback recovery. Administrator intervention is required.", 61);
	}

	return 0;
}



# Function : restart_send_redo_log
# Desc     : It is used to restart sending redo log to physical standby.
# Input    : $Owner - owner of executables
#          : $LD_ENV - Env var string returned from function "check_environment"
#          : $DBAMGR - client utility
#          : $CONNECTSTR - client utility specific connectstring
#          : $ResID - value of the dataguard resource
#          : $LinkRes - oracle resource name
#	   : $start_time - time when action started
# Returns  : 0 - Successfully started sending redo log
#          : 1 - Actiontimeout has expired without start sending redo log
#          : 2 - Re-try limit expired
#
sub restart_send_redo_log
{
        my ($Owner, $LD_ENV, $DBAMGR, $CONNECTSTR, $ResID, $LinkRes, $start_time) = @_;
        my ($SQLStmt, $sqlout, @sqloutarr);
	my ($DestID, $DestName, $Status, $waittm);

        my $ret = 0;
	my $i = 0;
        my ($actionto) = split('\n', `$HATYPE -value OraDG ActionTimeout`);

	$SQLStmt = "SELECT NAME FROM V\\\$PARAMETER2 WHERE VALUE LIKE '%service%';";
        $sqlout = &dglib::runsql($Owner,$LD_ENV,$DBAMGR,$CONNECTSTR,0,$SQLStmt);
        if ( $sqlout =~ /ORA-/i) {
        	VCSAG_LOG_MSG("E", "Unable to get archive destination log, error: $sqlout", 72, "$sqlout");
                exit 1;
        }

        @sqloutarr = split('\n', $sqlout);
        $DestName = uc($sqloutarr[4]);
        $SQLStmt = "SELECT DEST_ID FROM V\\\$ARCHIVE_DEST WHERE DEST_NAME='$DestName';";
        $sqlout = &dglib::runsql($Owner,$LD_ENV,$DBAMGR,$CONNECTSTR,0,$SQLStmt);
        if ( $sqlout =~ /ORA-/i) {
        	VCSAG_LOG_MSG("E", "Unable to get DEST_ID for archive destination log $DestName, error: $sqlout", 73, "DestName", "$sqlout");
                exit 1;
        }

	@sqloutarr = split('\n', $sqlout);
        $DestID = $sqloutarr[4];
	$DestID =~ s/^\s+//;

	my $DestStateName = "log_archive_dest_state_" . $DestID;

        $SQLStmt = "ALTER SYSTEM SET $DestStateName='RESET' SCOPE=BOTH;\n";
	$SQLStmt .= "ALTER SYSTEM SET $DestStateName='ENABLE' SCOPE=BOTH;\n";
        $sqlout = &dglib::runsql($Owner,$LD_ENV,$DBAMGR,$CONNECTSTR,0,$SQLStmt);
        if ( $sqlout =~ /ORA-/i) {
                VCSAG_LOG_MSG("E", "Unable to reset redo log connection, error: $sqlout", 69, "$sqlout");
                print "SQL error: $sqlout\n";
                exit 1;
        }

        my $cur_time = time();
        my $remaining_time = ($actionto - ($cur_time - $start_time));
        my $count = int ($remaining_time / 10) - 1;
	VCSAG_LOG_MSG("I", "value of count is $count", 81, "$count");
        if ( $count <= 0 ) {
		VCSAG_LOG_MSG("I", "No time left for proceeding for action entry point. Please increase actiontimeout. ActionTimeout value is too less to proceed for start sending redo log. Administrator need to increase the actiontimout and monitor interval.", 79);
		print "No time left for proceeding for action entry point. Please increase actiontimeout. Administrator need to increase the actiontimout and monitor interval.\n";
		exit 1;
        }

	$i = 0;
        while ( $i < $count ) {
                $SQLStmt = "SELECT STATUS FROM V\\\$ARCHIVE_DEST WHERE DEST_ID=$DestID;";
                $sqlout = &dglib::runsql($Owner,$LD_ENV,$DBAMGR,$CONNECTSTR,0,$SQLStmt);
                if ( $sqlout =~ /ERROR/i) {
	                VCSAG_LOG_MSG("I", "Sleeping for another 10 seconds as STATUS of archive dest for dest id $DestID is still in ERROR state.", 75, "$DestID");
                	$i++;
			$Status = 1;
                	sleep 10;
                } elsif ( $sqlout =~ /Valid/i ) {
                	VCSAG_LOG_MSG("I", "Status of archive dest for dest id $DestID is VALID: $sqlout", 77, "$DestID", "$sqlout");
			$Status = 0;
               		last;
                }
	}

	if ( ($i == $count) && ($Status) ) {
		return 2;
	}

        $SQLStmt = "ALTER SYSTEM ARCHIVE LOG CURRENT;";
        $sqlout = &dglib::runsql($Owner,$LD_ENV,$DBAMGR,$CONNECTSTR,0,$SQLStmt);
        if ( $sqlout =~ /ORA-/i) {
                VCSAG_LOG_MSG("E", "Unable to alter system archive log, error: $sqlout", 69, "$sqlout");
                print "SQL error: $sqlout\n";
                exit 1;
        }

	return 0;
}

# Function : get_current_rpo
# Desc     : Retrieve database rpo stats information from V$DATAGUARD_STAT
# Input    : $Owner - owner of executables
#          : $LD_ENV - Env var string returned from function "check_environment"
#          : $DBAMGR - client utility
#          : $CONNECTSTR - client utility specific connectstring
# Returns  : $sqlout
#
sub get_dataguard_stats_rpo() {
        my($Owner, $LD_ENV, $DBAMGR, $CONNECTSTR) = @_;
        my ($SQLStmt, $sqlout);
        my (@lines, $i);
	my $line;

        $SQLStmt = "COLUMN NAME FORMAT A24;\n";
        $SQLStmt .= "COLUMN VALUE FORMAT A16;\n";
        $SQLStmt .= "SELECT NAME, VALUE FROM V\\\$DATAGUARD_STATS;\n";

        $sqlout = `$SU $Owner -c "$LD_ENV; echo LD_LIBRARY_PATH - \\\$LD_LIBRARY_PATH; $DBAMGR" << !
$CONNECTSTR
$SQLStmt
!`;
	if ($sqlout =~ /\n(ORA-\d+.*$)/mi) {
		VCSAG_LOG_MSG("E", "Failed to get dataguard stats, error: $1", 100, "$1");
                print "failed to to get dataguard stats : $1\n";
                return 1;
        }

        @lines = split(/\n+/, $sqlout);
	shift(@lines);
	shift(@lines);
	shift(@lines);
	my $rpo = $lines[0];
	my @column = split (' ', $rpo);
	print "Replication RPO = $column[3] [hh:mm:ss]\n";
	print "\nRPO details reported by Dataguard:\n";
	while (@lines) {
		$line = shift @lines;
		if ($line =~ /estimated startup time/) {
			print "$line second(s)\n";
		} else {
			print "$line [hh:mm:ss]\n";
		}
	}
	return 0;
}
 
sub get_scn_rpo () {
	my ($ResID, $Owner, $LD_ENV, $DBAMGR, $CONNECTSTR) = @_;
	my ($SQLStmt, $sqlout);
	my ($RemClus, $RemNode);
	my $primary_SCN;
	my $standby_SCN;

	#Determine the SCN of STANDBY database.
	$SQLStmt = "SELECT TO_CHAR(CURRENT_SCN) FROM V\\\$DATABASE;\n";
	$sqlout = &dglib::runsql($Owner,$LD_ENV,$DBAMGR,$CONNECTSTR,0,$SQLStmt);

	if ($sqlout =~ /\n(ORA-\d+.*$)/mi) {
        	return 1;
        }

        my @lines = split(/\n+/, $sqlout);
        $standby_SCN = $lines[3];
        chomp($standby_SCN);
	
	my $ret = &dglib::get_remote_info($ResID, $RemClus, $RemNode);
	if ($ret) {
                        VCSAG_LOG_MSG("I","Unable to get Primary database information. Please check Primary database.", 55);
                        # Do nothing, primary database is down.
			return 1;
	}
        
	#set -sys and -clust attribute to work with RDC and gco
        if ($RemClus eq "" ) {
               	$RemoteNode = "-sys $RemNode";
        } else {
                $RemoteNode = "-sys $RemNode -clust $RemClus";
        } 
	
	# Determining the SCN of PRIMARY database.
	my $cmd = "$HARES -action $ResID getremoteSCN -actionargs CURRENT $RemoteNode 2>&1";
        my @SCNout = split ('\n', `$cmd`);
	$ret = $?;
	if ($ret != 0) {
		my $SCNout = join (" ", @SCNout);
		VCSSAG_LOG_MSG("E", "Unable to execute getremoteSCN action entry point on Remote cluster  error: $SCNout", 101, "$SCNout");
                print "getremoteSCN error: $SCNout\n";
                return 1;
        }
        $primary_SCN = $SCNout[1];

	my $scn_rpo = "Primary Database Current SCN $primary_SCN\n";
	$scn_rpo = $scn_rpo. "Standby Database Current SCN $standby_SCN";
	print "$scn_rpo";
	return 0;
}

sub offline_parent_sg {
	my ($ResID, $LocSys, $sCDBGrp) = @_;

	my ($sCommand, $sLocalSG);
	my @sCmdout;
	
	$sCommand = "$HARES -value $ResID Group";
	$sLocalSG = `$sCommand`;
	chomp($sLocalSG);
	
	$sCommand =  "$HAGRP -dep $sCDBGrp";
	@sCmdout = `$sCommand`;
	chomp(@sCmdout);
	
	foreach my $sSGdep (@sCmdout) {
		next if ($sSGdep =~ /^#/ || $sSGdep eq $sLocalSG);
		my $sPDBParent = (split(/\s+/, $sSGdep))[0];
		chomp($sPDBParent);
		my $sPDBGrpState = `$HAGRP -state $sPDBParent -sys $LocSys`;
		chomp($sPDBGrpState);
                if($sPDBGrpState =~ /ONLINE/i) {
                        VCSAG_LOG_MSG("I", "Trying to Offline $sPDBParent service group.", 105, "$sPDBParent");
                        system("$HAGRP -offline $sPDBParent -sys $LocSys");
                }
	}
}

1;
